﻿

namespace NRuler.Test
{
    using System;
    using System.Collections.Generic;
    using NRuler.Common;
    using NRuler.Conditions;
    using NRuler.Conflicts;
    using NRuler.Interfaces;
    using NRuler.Rete;
    using NRuler.Terms;
    using NUnit.Framework;
    using Test.Common;

    [TestFixture]
    public class BasicTest
    {
        #region WMETest

        [Test]
        [Category("RuleEngine")]
        public void WMETest_Add()
        {
            WME wme = new WME();
            wme.Fields[0] = "x";
            wme.Fields[1] = "y";
            wme.Fields[2] = "z";

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void WMETest_Remove()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }
                        
            Assert.AreEqual(rete.WorkingMemory.Count, 9);
            Assert.AreEqual(prod.Instances.Count, 1);

            // foamliu, 2008/11/24, remove w8 (B3 ^on table).
            rete.RemoveWme(wmes[7]);
            
            Assert.AreEqual(rete.WorkingMemory.Count, 8);
            Assert.AreEqual(prod.Instances.Count, 1);

            // foamliu, 2008/11/21, remove w5 (B2 ^left-of B3).
            rete.RemoveWme(wmes[4]);
           
            Assert.AreEqual(rete.WorkingMemory.Count, 7);
            Assert.AreEqual(prod.Instances.Count, 0);
        }

        #endregion

        #region TermTest

        [Test]
        [Category("RuleEngine")]
        public void TermTest_String()
        {
            WME wme = new WME(new StringTerm("12"), ">", new StringTerm("10"));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_String2()
        {
            WME wme = new WME('a', 'b', 'c');

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Integer()
        {
            WME wme = new WME(1, 2, 3);            

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Integer2()
        {
            IntegerTerm it1 = new IntegerTerm(250);
            IntegerTerm it2 = new IntegerTerm(251);

            Assert.IsFalse(it1.Equals(it2));
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Integer3()
        {
            IntegerTerm it1 = new IntegerTerm("100");
            IntegerTerm it2 = new IntegerTerm("100");

            Assert.IsTrue(it1.Equals(it2));
            Assert.IsTrue(it1.Value == 100);
            Assert.IsTrue(it1.Value == it2.Value);

            it2.Value = 101;
            Assert.IsFalse(it1.Equals(it2));
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Double()
        {
            WME wme = new WME(1.0d, 2.0d, 3.0d);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Double2()
        {
            DoubleTerm dt1 = new DoubleTerm(10.0d);
            DoubleTerm dt2 = new DoubleTerm(12.0d);

            Assert.IsFalse(dt1.Equals(dt2));
            
            dt1.Value = 12.0d;
            Assert.IsTrue(dt1.Equals(dt2));

        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Double3()
        {
            DoubleTerm dt1 = new DoubleTerm("10.0");
            DoubleTerm dt2 = new DoubleTerm("10.0");

            Assert.IsTrue(dt1.Equals(dt2));
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Double4()
        {
            DoubleTerm dt1 = new DoubleTerm(10.0m);
            DoubleTerm dt2 = new DoubleTerm(10.0m);

            Assert.AreEqual(dt1.Equals(dt2), true);            
        }

        //[Test]
        //[Category("RuleEngine")]
        //public void TermTest_Double5()
        //{
        //    DoubleTerm dt1 = new DoubleTerm(10.0f);
        //    DoubleTerm dt2 = new DoubleTerm(10.0f);

        //    Assert.AreEqual(dt1.Equals(dt2), true);
        //}

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Float()
        {
            WME wme = new WME(1.0f, 2.0f, 3.0f);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        /// <summary>
        /// decimal
        /// ±1.0 × 10e−28 to ±7.9 × 10e+28
        /// 28-29 significant digits 
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void TermTest_Decimal()
        {
            WME wme = new WME(1.0m, 2.0m, 3.0m);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Boolean()
        {
            WME wme = new WME(true, "!=", false);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Boolean2()
        {
            BooleanTerm bt1 = new BooleanTerm(false);
            BooleanTerm bt2 = new BooleanTerm(false);

            Assert.IsTrue(bt1.Equals(bt2));
            Assert.IsTrue(bt1.Value == false);

            bt2.Value = true;
            Assert.IsFalse(bt1.Equals(bt2));
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_DateTime()
        {
            // February 16, 1992, 12 hours, 15 minutes and 12 seconds.
            DateTime dt1 = DateTime.Parse("2/16/1992 12:15:12");
            // March 24, 2008, 2 hours, 45 minutes and 44 seconds.
            DateTime dt2 = DateTime.Parse("3/24/2008 02:45:44");

            WME wme = new WME(dt1, "<", dt2);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddWme(wme);
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_DateTime2()
        {
            // February 16, 1992, 12 hours, 15 minutes and 12 seconds.
            DateTimeTerm dt1 = new DateTimeTerm(DateTime.Parse("2/16/1992 12:15:12"));
            // March 24, 2008, 2 hours, 45 minutes and 44 seconds.
            DateTimeTerm dt2 = new DateTimeTerm(DateTime.Parse("3/24/2008 02:45:44"));

            Assert.IsFalse(dt1.Equals(dt2));

            dt2.Value = dt1.Value;
            Assert.IsTrue(dt1.Equals(dt2));
        }

        [Test]
        [Category("RuleEngine")]
        public void TermTest_Null()
        {
            NullTerm nt = new NullTerm();
            NullTerm nt2 = new NullTerm();
            nt2.Value = null;

            Assert.IsTrue(nt.Equals(null));
            Assert.IsTrue(nt.Value == null);
            Assert.IsTrue(nt.Equals(nt2));
            Assert.IsTrue(nt2.Equals(nt));
        }

        #endregion

        #region ProductionTest

        [Test]
        [Category("RuleEngine")]
        public void ProductionTest_Remove()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Assert.AreEqual(prod.Instances.Count, 1);

            // foamliu, 2008/11/21, remove whole production.
            rete.RemoveProduction(prod);
            
            Assert.AreEqual(prod.Instances.Count, 0);
        }

        #endregion

        #region ConditionTest

        /// <summary>
        /// This production matches if there is a stack of (at least) two blocks (designated by <x> and
        /// <y>) to the left of some block (designated by <z>) which is not both red and on some other
        /// block. Note that the occurrences of <z> inside the NCC refer to the binding of <z> from outside
        /// the NCC (in C2), while <w> is simply a new variable inside the NCC, since it does not occur
        /// anywhere in the (positive) conditions outside the NCC.
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Positive()
        {
            // 1. Add rules;
            // 2. Add facts.

            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);
            

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Assert.AreEqual(1, prod.Instances.Count);

            Instance inst = prod.Instances[0];

            Assert.AreEqual("B1", inst.GetVariableValue("x").ToString());
            Assert.AreEqual("B2", inst.GetVariableValue("y").ToString());
            Assert.AreEqual("B3", inst.GetVariableValue("z").ToString());

        }

        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Positive2()
        {
            // 1. Add facts;
            // 2. Add rules.


            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));

            ReteInferenceEngine rete = new ReteInferenceEngine();

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            rete.AddProduction(prod);

            Assert.AreEqual(prod.Instances.Count, 1);


        }

        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Positive3()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));

            ReteInferenceEngine rete = new ReteInferenceEngine();

            rete.AddWme(wmes[0]);
            rete.AddWme(wmes[1]);
            rete.AddWme(wmes[2]);
            rete.AddWme(wmes[3]);
            //rete.AddWme(wmes[4]);
            //"B2", "^left-of", "B3"
            rete.AddWme(wmes[5]);
            rete.AddWme(wmes[6]);
            rete.AddWme(wmes[7]);
            rete.AddWme(wmes[8]);

            rete.AddProduction(prod);

            Assert.AreEqual(prod.Instances.Count, 0);

            rete.AddWme(wmes[4]);

            Assert.AreEqual(prod.Instances.Count, 1);

            
        }

        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Positive4()
        {
            // 1. Add rules;
            // 2. Add facts.

            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";            
            
            // foamliu, 2008/12/09, ordering does not matter.
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);


            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Assert.AreEqual(1, prod.Instances.Count);

            Instance inst = prod.Instances[0];

            Assert.AreEqual("B1", inst.GetVariableValue("x").ToString());
            Assert.AreEqual("B2", inst.GetVariableValue("y").ToString());
            Assert.AreEqual("B3", inst.GetVariableValue("z").ToString());

        }

        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Negative()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));

            NegativeCondition nc = new NegativeCondition(/*"C3", */z, "^color", "red");
            prod.LHS.Add(nc);

            nc.Label = "C3";
            Assert.AreEqual(nc.Label, "C3");

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Assert.AreEqual(1, prod.Instances.Count);

            Instance inst = prod.Instances[0];

            Assert.AreEqual("B1", inst.GetVariableValue("x").ToString());
            Assert.AreEqual("B3", inst.GetVariableValue("y").ToString());
            Assert.AreEqual("B4", inst.GetVariableValue("z").ToString());
        }

        /// <summary>
        /// foamliu, 2008/12/10.
        /// 
        /// Please note, negative conditon:
        ///     (1) only one;
        ///     (2) must be last.
        ///     
        /// This is a negative test for negative conditon, :)
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Negative2()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            
            // foamliu, 2008/12/10, negative condition must be the LAST ONE.     
            //
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new NegativeCondition("C3", z, "^color", "red"));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));            

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Assert.AreEqual(0, prod.Instances.Count);
            
        }

        /// <summary>
        /// foamliu, 2008/12/10.
        /// 
        /// Please note, negative conditon:
        ///     (1) only one;
        ///     (2) must be last.
        ///     
        /// This is a negative test for negative conditon, :)
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Negative3()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();            

            // foamliu, 2008/12/10, two negative conditions will not work.
            // 
            prod.LHS.Add(new NegativeCondition("C1", x, "^on", y));
            prod.LHS.Add(new NegativeCondition("C2", z, "^color", "red"));
            prod.LHS.Add(new PositiveCondition("C3", y, "^left-of", z));

            // 查询翻译成自然语言为： 找这样的x, y, z, 使得: 
            //  not x ^on y and not z ^color red and y ^left-of z
            //  先看y ^left-of z, 显然只有两种情况：
            //      (1). y=B2, z=B3;
            //      (2). y=B3, z=B4.
            //  对于(1), 因为 有w9 所以可以排除.
            //  对于(2), 因为 有w5 所以也可以排除.
            //  所以如果正确的到, 应该得到0个结果, 然而由于NegativeCondition只能有一个,而且只能放在最后,
            //  使得这种表达式无法work,于是也就可能不是0了.

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            // 但是没想到却是0, 只能说可对可错, 却没法断定一定出错, :(.
            Assert.AreEqual(0, prod.Instances.Count);

        }

        /// <summary>
        /// foamliu, 2008/12/10.
        /// only negative condition can yield nothing
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Negative4()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            // foamliu, 2008/12/09, only negative condition can yield nothing
            //
            prod.LHS.Add(new NegativeCondition("C1", x, "^on", y));            

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Assert.AreEqual(0, prod.Instances.Count);

        }

        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_NCC()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");
            Variable w = new Variable("w");

            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));

            NCCCondition neg = new NCCCondition(z, "^color", "red");
            //neg.SubConditions.Add(new PositiveCondition("C3", z, "^color", "red"));
            neg.Label = "C3";
            neg.SubConditions.Add(new PositiveCondition("C4", z, "^on", w));
            prod.LHS.Add(neg);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            Assert.AreEqual(1, prod.Instances.Count);

            Instance inst = prod.Instances[0];

            Assert.AreEqual("B1", inst.GetVariableValue("x").ToString());
            Assert.AreEqual("B3", inst.GetVariableValue("y").ToString());
            Assert.AreEqual("B4", inst.GetVariableValue("z").ToString());
        }

        /// <summary>
        /// foamliu, 2008/12/10, NCC replace negative condition in ConditionTest_Negative
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_NCC2()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Production prod = new Production();
            
            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");            

            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));

            NCCCondition neg = new NCCCondition("C3", z, "^color", "red");
            //neg.SubConditions.Add(new PositiveCondition("C3", z, "^color", "red"));            
            prod.LHS.Add(neg);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            Assert.AreEqual(1, prod.Instances.Count);

            Instance inst = prod.Instances[0];

            Assert.AreEqual("B1", inst.GetVariableValue("x").ToString());
            Assert.AreEqual("B3", inst.GetVariableValue("y").ToString());
            Assert.AreEqual("B4", inst.GetVariableValue("z").ToString());
        }

        /// <summary>
        /// foamliu, 2008/12/10, does ordering matter?
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_NCC3()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // 查询:
            // x ^on y and not z ^color red and y ^left z
            //  先看y ^left-of z, 显然只有两种情况：
            //      (1). y=B2, z=B3;
            //      (2). y=B3, z=B4.
            //  对于(1), 因为 有w9 所以可以排除.
            //  对于(2), 因为 有w2 所以成立.

            // 所以有满足条件的一组解:
            // x=B1, y=B3, z=B4.

            Production prod = new Production();

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));

            NCCCondition neg = new NCCCondition();
            neg.SubConditions.Add(new PositiveCondition("C2", z, "^color", "red"));
            prod.LHS.Add(neg);

            prod.LHS.Add(new PositiveCondition("C3", y, "^left-of", z));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            Assert.AreEqual(1, prod.Instances.Count);

            Instance inst = prod.Instances[0];

            Assert.AreEqual("B1", inst.GetVariableValue("x").ToString());
            Assert.AreEqual("B3", inst.GetVariableValue("y").ToString());
            Assert.AreEqual("B4", inst.GetVariableValue("z").ToString());
        }

        /// <summary>
        /// "exist" is just syntax sugar 
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Exist()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // exist x,y,z, s.t. <x> ^on <y> && <y> ^left-of <z> && <z> ^color red
            // instances > 0
            
            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "red"));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            Assert.Less(0, prod.Instances.Count);            
        }

        /// <summary>
        /// "exist" is just syntax sugar 
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_NotExist()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // exist x,y,z, s.t. <x> ^on <y> && <y> ^left-of <z> && <z> ^color red
            // instances > 0

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.Name = "find-stack-of-two-blocks-to-the-left-of-a-red-block";
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));
            prod.LHS.Add(new PositiveCondition("C2", y, "^left-of", z));
            prod.LHS.Add(new PositiveCondition("C3", z, "^color", "blue"));

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            Assert.AreEqual(0, prod.Instances.Count);
        }

        /// <summary>
        /// "forall" is just syntax sugar 
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_ForAll()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // for all x, <x> ^on <y> -> <y> ^left-of <z> "and" <z> ^color red
            // not exists: <x> ^on <y> "and" not (<y> ^left-of <z> "and" <z> ^color red)

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.LHS.Add(new PositiveCondition("C1", x, "^on", y));            

            NCCCondition neg = new NCCCondition();
            neg.SubConditions.Add(new PositiveCondition("C2", y, "^left-of", z));
            neg.SubConditions.Add(new PositiveCondition("C3", z, "^color", "red"));
            prod.LHS.Add(neg);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            // for-all is incorrect, exists.
            Assert.Less(0, prod.Instances.Count);
            
        }

        /// <summary>
        /// every red block has a blue block on top of it
        /// there is no red block that does not have a blue block on top of it
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_ForAll2()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // for all x, F1: <x> ^color red -> exists y. s.t. <y> ^on <x> "and" <y> ^color blue
            // not exists: F2: <x> ^color red "and" not (<y> ^on <x> "and" <y> ^color blue)
            // instance number for F2 is 0
            // instance number for !F2 is != 0

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();

            prod.LHS.Add(new PositiveCondition("C1", x, "^color", "red"));
            NCCCondition neg = new NCCCondition();
            neg.SubConditions.Add(new PositiveCondition("C2", y, "^on", x));
            neg.SubConditions.Add(new PositiveCondition("C3", y, "^color", "blue"));            
            prod.LHS.Add(neg);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            // for-all is incorrect, exists.
            // for example, B1 is red, but it does not have a y, y on B1 and y is blue.
            //
            Assert.Less(0, prod.Instances.Count);

        }

        /// <summary>
        /// every red block has a blue block on top of it
        /// there is no red block that does not have a blue block on top of it
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_ForAll3()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // for all x, F1: <x> ^color red -> exists y. s.t. <y> ^on <x> "and" <y> ^color blue
            // not exists: F2: <x> ^color red "and" not (<y> ^on <x> "and" <y> ^color blue)
            // instance number for F2 is 0
            // instance number for !F2 is != 0

            // 还是用中文吧 :(
            // 比如命题 P1说： 对任何x， <x> ^color red 则必然有: <y> ^on <x> "and" <y> ^color blue
            // P1可以翻译成等价命题： 不存在x， <x> ^color red 且 not (<y> ^on <x> "and" <y> ^color blue)
            // 也就是如下命题进行query得到match数应该为0：
            // <x> ^color red "and" not (<y> ^on <x> "and" <y> ^color blue)

            // 而从所列事实可见，此命题为假，则上述query得到的match数应该不为0.

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();
            prod.LHS.Add(new PositiveCondition("C1", x, "^color", "red"));
            
            NCCCondition neg = new NCCCondition();
            neg.SubConditions.Add(new PositiveCondition("C2", y, "^on", x));
            neg.SubConditions.Add(new PositiveCondition("C3", y, "^color", "blue"));
            prod.LHS.Add(neg);            

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            // for-all is incorrect, exists.
            // for example, B1 is red, but it does not have a y, y on B1 and y is blue.
            // 
            Assert.AreNotEqual(0, prod.Instances.Count);

        }

        /// <summary>
        /// "or" is just syntax sugar 
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Or()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // B1 ^on B2 "or" B2 ^on B3
            // not (not(B1 ^on B2) "and" not(B2 ^on B3))

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();           

            NCCCondition neg = new NCCCondition();
            NCCCondition neg1 = new NCCCondition();
            NCCCondition neg2 = new NCCCondition();
            neg.SubConditions.Add(neg1);
            neg.SubConditions.Add(neg2);
            neg1.SubConditions.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            neg1.SubConditions.Add(new PositiveCondition("C2", "B2", "^on", "B3"));            
            prod.LHS.Add(neg);

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            // foamliu, 2008/12/10, B1 ^on B2 "or" B2 ^on B3 should be true.
            // because B1 ^on B2 is true
            //
            Assert.AreNotEqual(0, prod.Instances.Count);            
        }

        /// <summary>
        /// for compare, this time it is and
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void ConditionTest_Or2()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            // B1 ^on B2 "or" B2 ^on B3
            // not (not(B1 ^on B2) "and" not(B2 ^on B3))

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod = new Production();

            prod.LHS.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            prod.LHS.Add(new PositiveCondition("C2", "B2", "^on", "B3"));            

            ReteInferenceEngine rete = new ReteInferenceEngine();
            rete.AddProduction(prod);

            foreach (WME wme in wmes)
            {
                rete.AddWme(wme);
            }

            // foamliu, 2008/12/10, B1 ^on B2 "and" B2 ^on B3 should be false.
            // because B2 ^on B3 is false (CWA: close world assumption)
            //
            Assert.AreEqual(0, prod.Instances.Count);
        }

        #endregion
        
        #region AgendaTest

        [Test]
        [Category("RuleEngine")]
        public void AgendaTest_Basic()
        {
            WME[] wmes = new WME[] {
                new WME("B1", "^on", "B2"),     // w1
                new WME("B1", "^on", "B3"),     // w2
                new WME("B1", "^color", "red"), // w3
                new WME("B2", "^on", "table"),  // w4
                //new WME("B2", "^left-of", "B3"),// w5
                new WME("B2", "^color", "blue"),// w6
                new WME("B3", "^left-of", "B4"),// w7
                new WME("B3", "^on", "table"),  // w8
                new WME("B3", "^color", "red"), // w9                      
            };

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            Production prod1 = new Production();
            prod1.LHS.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            prod1.RHS.Add(new AssertCondition("B2", "^left-of", "B3"));

            Production prod2 = new Production();
            prod2.LHS.Add(new PositiveCondition("C2", x, "^on", y));
            prod2.LHS.Add(new PositiveCondition("C3", y, "^left-of", z));
            prod2.LHS.Add(new PositiveCondition("C4", z, "^color", "red"));

            AssertCondition rhs = new AssertCondition("C5", x, "is", "on top");
            prod2.RHS.Add(rhs);

            Agenda agenda = new Agenda();
            agenda.Name = "agenda1";
            agenda.ConflictResolutionStrategy = new BasicResolver();
            agenda.AddProduction(prod1);
            agenda.AddProduction(prod2);
            agenda.AddFacts(wmes);
            
            agenda.Run();
            
            Assert.AreEqual(prod2.Instances.Count, 1);

        }

        [Test]
        [Category("RuleEngine")]
        public void AgendaTest_Basic2()
        {
            Production prod1 = new Production();
            prod1.Priority = 0;
            prod1.LHS.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            prod1.RHS.Add(new AssertCondition("B2", "^left-of", "B3"));

            Production prod2 = new Production();
            prod2.Priority = 100;
            prod2.LHS.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            prod2.RHS.Add(new RetractCondition("C2", "B1", "^on", "B2"));

            Agenda agenda = new Agenda();
            agenda.Name = "agenda2";
            agenda.ConflictResolutionStrategy = new BasicResolver();
            agenda.AddProduction(prod1);
            agenda.AddProduction(prod2);
            agenda.AddFact(new WME("B1", "^on", "B2"));
            agenda.Run();

            Assert.AreEqual(prod1.Instances.Count, 1);
            Assert.AreEqual(prod2.Instances.Count, 1);
            Assert.AreEqual(agenda.ActivatedRuleCount, 2);
        }

        [Test]
        [Category("RuleEngine")]
        public void AgendaTest_Basic3()
        {
            Production prod1 = new Production();
            prod1.Priority = 0;
            prod1.LHS.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            prod1.RHS.Add(new AssertCondition("B2", "^left-of", "B3"));

            Production prod2 = new Production();
            prod2.Priority = 100;
            prod2.LHS.Add(new PositiveCondition("C1", "B1", "^on", "B2"));
            prod2.RHS.Add(new RetractCondition("C2", "B1", "^on", "B2"));

            Agenda agenda = new Agenda();
            agenda.Name = "agenda3";
            agenda.ConflictResolutionStrategy = new BasicResolver();
            agenda.AddProduction(prod1);
            agenda.AddProduction(prod2);
            agenda.AddFact(new WME("B1", "^on", "B2"));
            agenda.Run();

            Assert.AreEqual(prod1.Instances.Count, 1);
            Assert.AreEqual(prod2.Instances.Count, 1);
            Assert.AreEqual(agenda.ActivatedRuleCount, 2);
        }



        #endregion

        #region PrologTest

        [Test]
        [Category("RuleEngine")]
        public void FunctionTest_GrandParents()
        {
            ReteInferenceEngine rete = new ReteInferenceEngine();

            Production prod = new Production();
            prod.Name = "Whose your daddy";

            Variable x = new Variable("x");
            Variable y = new Variable("y");
            Variable z = new Variable("z");
            prod.LHS.Add(new PositiveCondition(x, "parent", y));

            rete.AddProduction(prod);

            Production prod1 = new Production();
            prod1.Name = "Whose your grand daddy";

            prod1.LHS.Add(new PositiveCondition(x, "parent", y));
            prod1.LHS.Add(new PositiveCondition(y, "parent", z));

            rete.AddProduction(prod1);

            Production prod2 = new Production();
            prod2.Name = "Whose your grand child";

            prod2.LHS.Add(new PositiveCondition(x, "gparent", y));

            rete.AddProduction(prod2);

            Production prod3 = new Production();
            prod3.Name = "Whose your great grand daddy";

            prod3.LHS.Add(new PositiveCondition(x, "parent", y));
            prod3.LHS.Add(new PositiveCondition(y, "gparent", z));

            rete.AddProduction(prod3);

            List<WME> wmes = new List<WME>();

            wmes.Add(new WME("george", "gchild", "kendall"));
            wmes.Add(new WME("kendall", "parent", "immanuel"));
            wmes.Add(new WME("louis", "gparent", "richard"));
            wmes.Add(new WME("jacques", "parent", "jean-francois"));
            wmes.Add(new WME("friedrich", "parent", "louis"));
            wmes.Add(new WME("louis", "gparent", "willard"));
            wmes.Add(new WME("willard", "gchild", "ludwig"));
            wmes.Add(new WME("miles", "parent", "ludwig"));
            wmes.Add(new WME("ludwig", "gchild", "noam"));
            wmes.Add(new WME("miles", "gparent", "wilhelm"));
            wmes.Add(new WME("friedrich", "parent", "jean-francois"));
            wmes.Add(new WME("bijan", "gchild", "george"));
            wmes.Add(new WME("wilhelm", "gparent", "bertrand"));
            wmes.Add(new WME("alan", "parent", "bijan"));
            wmes.Add(new WME("kendall", "parent", "gottlob"));
            wmes.Add(new WME("david", "gchild", "bertrand"));
            wmes.Add(new WME("alan", "gchild", "henry"));
            wmes.Add(new WME("kendall", "gparent", "louis"));
            wmes.Add(new WME("willard", "gchild", "willard"));
            wmes.Add(new WME("richard", "gchild", "pat"));
            wmes.Add(new WME("miles", "parent", "friedrich"));
            wmes.Add(new WME("noam", "gparent", "ludwig"));
            wmes.Add(new WME("jean-francois", "gchild", "bertrand"));
            wmes.Add(new WME("louis", "parent", "bijan"));
            wmes.Add(new WME("willard", "parent", "kendall"));
            wmes.Add(new WME("gottlob", "gparent", "kendall"));
            wmes.Add(new WME("immanuel", "gchild", "wilhelm"));
            wmes.Add(new WME("george", "parent", "drew"));
            wmes.Add(new WME("friedrich", "gchild", "david"));
            wmes.Add(new WME("gottlob", "gchild", "bertrand"));
            wmes.Add(new WME("wilhelm", "parent", "ludwig"));
            wmes.Add(new WME("henry", "gchild", "willard"));
            wmes.Add(new WME("alan", "gchild", "richard"));
            wmes.Add(new WME("george", "gchild", "miles"));
            wmes.Add(new WME("george", "gparent", "willard"));
            wmes.Add(new WME("alasdair", "gparent", "willard"));
            wmes.Add(new WME("willard", "gparent", "immanuel"));
            wmes.Add(new WME("jacques", "gparent", "george"));
            wmes.Add(new WME("henry", "gchild", "rudolf"));
            wmes.Add(new WME("wilhelm", "gparent", "miles"));
            wmes.Add(new WME("noam", "gparent", "jean-francois"));
            wmes.Add(new WME("pat", "gchild", "friedrich"));
            wmes.Add(new WME("rudolf", "gchild", "david"));
            wmes.Add(new WME("john", "gchild", "stanley"));
            wmes.Add(new WME("jacques", "gchild", "bijan"));
            wmes.Add(new WME("george", "parent", "miles"));
            wmes.Add(new WME("louis", "parent", "drew"));
            wmes.Add(new WME("rudolf", "parent", "jacques"));
            wmes.Add(new WME("bertrand", "gparent", "ludwig"));
            wmes.Add(new WME("gottlob", "parent", "john"));
            wmes.Add(new WME("miles", "gchild", "miles"));
            wmes.Add(new WME("bijan", "gchild", "noam"));
            wmes.Add(new WME("jean-francois", "gparent", "george"));
            wmes.Add(new WME("miles", "gparent", "henry"));
            wmes.Add(new WME("kendall", "gchild", "ludwig"));
            wmes.Add(new WME("louis", "gparent", "david"));
            wmes.Add(new WME("noam", "parent", "george"));
            wmes.Add(new WME("pat", "parent", "richard"));
            wmes.Add(new WME("george", "parent", "ludwig"));
            wmes.Add(new WME("george", "gparent", "alan"));
            wmes.Add(new WME("pat", "gparent", "louis"));
            wmes.Add(new WME("david", "parent", "friedrich"));
            wmes.Add(new WME("wilhelm", "gchild", "alasdair"));
            wmes.Add(new WME("pat", "gchild", "bertrand"));
            wmes.Add(new WME("noam", "parent", "stanley"));
            wmes.Add(new WME("friedrich", "gparent", "bijan"));
            wmes.Add(new WME("willard", "gchild", "bertrand"));
            wmes.Add(new WME("kendall", "gparent", "noam"));
            wmes.Add(new WME("kendall", "gparent", "george"));
            wmes.Add(new WME("drew", "parent", "gottlob"));
            wmes.Add(new WME("david", "parent", "kendall"));
            wmes.Add(new WME("alan", "gparent", "gottlob"));
            wmes.Add(new WME("richard", "parent", "jacques"));
            wmes.Add(new WME("alasdair", "gchild", "david"));
            wmes.Add(new WME("richard", "parent", "bijan"));
            wmes.Add(new WME("jean-francois", "gparent", "alan"));
            wmes.Add(new WME("jacques", "gchild", "louis"));
            wmes.Add(new WME("miles", "gchild", "jean-francois"));
            wmes.Add(new WME("pat", "parent", "miles"));
            wmes.Add(new WME("kendall", "parent", "gottlob"));
            wmes.Add(new WME("gottlob", "parent", "willard"));
            wmes.Add(new WME("jean-francois", "parent", "willard"));
            wmes.Add(new WME("stanley", "gchild", "ludwig"));
            wmes.Add(new WME("john", "gchild", "miles"));
            wmes.Add(new WME("john", "parent", "jean-francois"));
            wmes.Add(new WME("alan", "gchild", "louis"));
            wmes.Add(new WME("stanley", "gchild", "wilhelm"));
            wmes.Add(new WME("drew", "gparent", "miles"));
            wmes.Add(new WME("jacques", "parent", "immanuel"));
            wmes.Add(new WME("david", "gparent", "willard"));
            wmes.Add(new WME("miles", "parent", "alasdair"));
            wmes.Add(new WME("immanuel", "gparent", "ludwig"));
            wmes.Add(new WME("kendall", "gparent", "henry"));
            wmes.Add(new WME("immanuel", "gparent", "ludwig"));
            wmes.Add(new WME("miles", "gchild", "john"));
            wmes.Add(new WME("louis", "gchild", "alan"));
            wmes.Add(new WME("willard", "parent", "alan"));
            wmes.Add(new WME("kendall", "parent", "ludwig"));
            wmes.Add(new WME("george", "gchild", "alasdair"));
            wmes.Add(new WME("richard", "gparent", "bertrand"));

            int i = 1;
            foreach (WME wme in wmes)
            {
                wme.Name = "W" + i++;
                rete.AddWme(wme);
            }

            // foamliu, 2008/11/21, why not 100, may be two dups, need to investigate later.
            //
            Assert.AreEqual(rete.WorkingMemory.Count, 98);
            //Assert.AreEqual(prod.P_Node.Items.Count, 33);
            Assert.AreEqual(prod.Instances.Count, 33);
            //Assert.AreEqual(prod1.P_Node.Items.Count, 42);
            Assert.AreEqual(prod1.Instances.Count, 42);
            //Assert.AreEqual(prod2.P_Node.Items.Count, 29);
            Assert.AreEqual(prod2.Instances.Count, 29);
            //Assert.AreEqual(prod3.P_Node.Items.Count, 38);
            Assert.AreEqual(prod3.Instances.Count, 38);

        }

        /// <summary>
        /// http://www.csupomona.edu/~jrfisher/www/prolog_tutorial/2_1.html
        /// </summary>
        [Test]
        [Category("RuleEngine")]
        public void PrologTest_MapColorings ()
        {
            // 1. Add rules;
            // 2. Add facts.

            // adjacent(1,2).         adjacent(2,1). 
            // adjacent(1,3).         adjacent(3,1). 
            // adjacent(1,4).         adjacent(4,1). 
            // adjacent(1,5).         adjacent(5,1). 
            // adjacent(2,3).         adjacent(3,2). 
            // adjacent(2,4).         adjacent(4,2). 
            // adjacent(3,4).         adjacent(4,3). 
            // adjacent(4,5).         adjacent(5,4).

            WME[] wmes = new WME[] {
                new WME(1, "^adjacent", 2),
                new WME(2, "^adjacent", 1), 
                new WME(1, "^adjacent", 3),
                new WME(3, "^adjacent", 1),
                new WME(1, "^adjacent", 4),
                new WME(4, "^adjacent", 1),
                new WME(1, "^adjacent", 5),
                new WME(5, "^adjacent", 1),
                new WME(2, "^adjacent", 3),        
                new WME(3, "^adjacent", 2),  
                new WME(2, "^adjacent", 4),  
                new WME(4, "^adjacent", 2),  
                new WME(3, "^adjacent", 4),  
                new WME(4, "^adjacent", 3),  
                new WME(4, "^adjacent", 5),  
                new WME(5, "^adjacent", 4),  
            };

            ReteInferenceEngine rete = new ReteInferenceEngine();            

            foreach (WME w in wmes)
            {
                rete.AddWme(w);
            }

            Variable R = new Variable("R");
            Variable y = new Variable("y");
            Variable z = new Variable("z");

            // ?- adjacent(2,3). 
            Production prod1 = new Production();
            prod1.LHS.Add(new PositiveCondition(2, "^adjacent", 3));
            rete.AddProduction(prod1);
            Assert.AreEqual(1, prod1.Instances.Count);

            // ?- adjacent(5,3). 
            Production prod2 = new Production();
            prod2.LHS.Add(new PositiveCondition(5, "^adjacent", 3));
            rete.AddProduction(prod2);
            Assert.AreEqual(0, prod2.Instances.Count);

            // ?- adjacent(3,R). 
            Production prod3 = new Production();
            prod3.LHS.Add(new PositiveCondition(3, "^adjacent", R));
            rete.AddProduction(prod3);
            Assert.AreEqual(3, prod3.Instances.Count);           

            Assert.AreEqual("1", prod3.Instances[0].GetVariableValue("R").ToString());
            Assert.AreEqual("2", prod3.Instances[1].GetVariableValue("R").ToString());
            Assert.AreEqual("4", prod3.Instances[2].GetVariableValue("R").ToString());



        }

        #endregion

        #region Interfaces

        [Test]
        public void Interfaces_RuleSetParserTest()
        {
            string xml = "<rule-set><rule name='age30'><parameter identifier ='person'><class>UnitTests.Person</class></parameter><condition>person.Age == 30</condition><consequence>person.SayHello();</consequence></rule></rule-set>";
            RuleSet ruleSet = RuleSet.LoadXml(xml);
            Assert.AreEqual(1, ruleSet.Count);

            Rule rule = ruleSet[0];
            Assert.AreEqual("age30", rule.Name);
            Assert.AreEqual("person", rule.ParameterList[0].Identifier);
            Assert.AreEqual("UnitTests.Person", rule.ParameterList[0].Class);
            Assert.AreEqual("person.Age == 30", rule.ConditionList[0].CodeSnippet);
            Assert.AreEqual("person.SayHello();", rule.Consequence.CodeSnippet);

            Assert.AreEqual(xml, ruleSet.ToXmlString());
        }        

        [Test]
        public void Interfaces_WorkingMemoryTest()
        {
            Person person = new Person();
            person.Name = "YangLiu";
            person.Age = 30;           

            WorkingMemory wm = new WorkingMemory();
            wm.AssertObject(person, "person");

            //List<WME> facts = wm.Agenda.InitialFacts;
            //foreach (WME fact in facts)
            //{
            //    Logger.Info(fact.ToString());
            //}

            Assert.AreEqual(2, wm.Agenda.InitialFacts.Count);
            Assert.AreEqual("person ^Name YangLiu", wm.Agenda.InitialFacts[0].ToString());
            Assert.AreEqual("person ^Age 30", wm.Agenda.InitialFacts[1].ToString());
        }

        [Test]
        public void Interfaces_RuleSetTest()
        {
            string xml = "<rule-set><rule name='age30'><parameter identifier ='person'><class>NRuler.Test.Common.Person</class></parameter><condition>person.Age == 30</condition><consequence>person.SayHello();</consequence></rule></rule-set>";
            RuleSet ruleSet = RuleSet.LoadXml(xml);

        }

        [Test]
        public void Interfaces_ConsequenceInvokerTest()
        {
            Person person = new Person();
            person.Name = "YangLiu";
            person.Age = 30;
            ApplicationData appData = new ApplicationData();
            appData.Add("person", person);

            string xml = "<rule-set><rule name='age30'><parameter identifier ='person'><class>NRuler.Test.Common.Person</class></parameter><condition>person.Age == 30</condition><consequence>person.SayHello();</consequence></rule></rule-set>";
            RuleSet ruleSet = RuleSet.LoadXml(xml);
            ruleSet[0].AppData = appData;

            ConsequenceInvoker invoker = new ConsequenceInvoker(ruleSet[0]);
            invoker.Invoke();
        }

        [Test]
        public void Interfaces_FireRuleTest()
        {
            Person person = new Person();
            person.Name = "YangLiu";
            person.Age = 30;   
            ApplicationData appData = new ApplicationData();
            appData.Add("person", person);

            string xml = "<rule-set><rule name='age30'><parameter identifier ='person'><class>NRuler.Test.Common.Person</class></parameter><condition>person.Age == 30</condition><consequence>person.SayHello();</consequence></rule></rule-set>";
            RuleSet ruleSet = RuleSet.LoadXml(xml);

            Logger.Info("First firing started.");
            // 第一次
            foreach (Rule r in ruleSet.List)
            {
                r.AppData = appData;
                r.Fire();
            }
            Logger.Info("Second firing started.");
            // 第二次, 应该比第一次快很多, 所谓 JIT 嘛!
            foreach (Rule r in ruleSet.List)
            {
                r.AppData = appData;
                r.Fire();
            }
            Logger.Info("Second firing completed.");

        }

        #endregion
    }

    
}

